<?php
// $Id: OoxLock.php,v 1.1.4.1 2010/11/27 15:36:09 pounard Exp $

/**
 * Exception thrown when aptempting to acquire an already locked lock.
 */
class LockAlreadyAcquiredException extends OoxException
{
  /**
   * Lock database row.
   * 
   * @var object
   */
  protected $_lock;

  /**
   * Current user static cache, since Drupal does not do cache into the
   * user_load() function.
   */
  protected $_user;

  /**
   * Get user which owns the current lock.
   * 
   * @return object
   *   Drupal database account row.
   */
  public function getAccount() {
    if (!isset($this->_user)) {
      if ($this->_lock->uid) {
        $this->_user = user_load($this->_lock->uid);
      }
      // Create a fake user, with a non existing uid, which name is
      // the current translation of "System".
      else {
        $this->_user = new stdClass();
        $this->_user->uid = -1;
        $this->_user->name = t("System");
        $this->_user->mail = variable_get('site_mail', '');
      }
    }
    return $this->_user;
  }

  /**
   * Get the lock database row.
   */
  public function getLock() {
    return $this->_lock;
  }

  /**
   * Build instance.
   * 
   * @param object $lock
   *   Lock database row.
   */
  public function __construct($lock) {
    parent::__construct("Already locked lock");
    $this->_lock = $lock;
  }
}

/**
 * Simple semaphore based persistant locking framework. This interface defines
 * the lock backend API.
 * 
 * Locks are acquired on a name basis. Each acquired lock will raise exception
 * on new acquire attempts until it's being either pragmaticaly released, or
 * when its expiration time comes out of date.
 * 
 * For each lock, you can also give a token value. Any locking attempt of the
 * exact same lock with the exact same token will return sucessful.
 * 
 * If user is currently logged, each semaphore can store its uid. When done,
 * all current active locks will be released on user session destroy.
 * 
 * A default backend using a database table for locking is given, and used per
 * default.
 * 
 * @see OoxLock
 * @see oox_user()
 */
interface ILock
{
  const DEFAULT_GROUP = 'default';

  /**
   * Acquire a lock.
   * 
   * @param string $name
   *   Lock name.
   * @param int $lifetime = LIFETIME_DEFAULT
   *   (optional) Lock lifetime (in seconds).
   * @param string $group = ILock::DEFAULT_GROUP
   *   (optional) A group name for lock, this will allow to remove all locks for
   *   a certain group.
   * @param boolean $forUser = FALSE
   *   (optional) If set to TRUE, the current lock will be linked to the current
   *   logged in user (except for anonymous). This means that on session destroy
   *   if no other session is active, the lock will be removed.
   * @param string $token = NULL
   *   (optional) If set, set a token the lock. Tokens will allow more than one
   *   locking attempt on the same lock, using the same token. It can also allow
   *   lock sharing among users.
   * 
   * @throws LockAlreadyAcquiredException
   *   If lock already been acquired by another thread.
   */
  public function acquire($name, $lifetime = ILock::LIFETIME_DEFAULT, $group = OoxLock::DEFAULT_GROUP, $forUser = FALSE, $token = NULL);

  /**
   * Release a lock.
   * If lock is already free or does not exists, remains silent.
   * 
   * @param string $name
   *   Lock name.
   */
  public function release($name);

  /**
   * Get alive lock.
   * 
   * @param string $name
   *   Lock name.
   * @param string $group = NULL
   *   (optional) If set, will attempt to find by matching group.
   * @param int $uid = NULL
   *   (optional) If set, will attempt to find by matching uid.
   * @param string $token
   *   (optional) If set, will attempt to find by matching token.
   * 
   * @return object
   *   Lock database object if locked, FALSE else.
   */
  public function getLock($name, $group = NULL, $uid = NULL, $token = NULL);

  /**
   * Clear outdated locks.
   * 
   * Both of the following parameters can be used at once.
   * 
   * @param int $uid = 0
   *   (optional) If set, this will only wipe out all locks for the given
   *   user identifier.
   * @param string $group = NULL
   *   (optional) If set, this will wipe out all locks for the given group.
   * @param string $token = NULL
   *   (optional) If set, this will wipe out all locks for the given token.
   */
  public function wipeOutLocks($uid = 0, $group = NULL, $token = NULL);

  /**
   * Some cache backend can not run on some environment. This function will be
   * called at singleton instanciation time, if it returns FALSE, the singleton
   * instanciation will drop to the default backend as failsafe choice.
   * 
   * Remember that this check will done at each page request, it is your job to
   * ensure it won't be ressource intensive.
   * 
   * @return boolean
   */
  public function checkEnvironment();
}

/**
 * Default implementation, and singleton implementation. Fine for most usages.
 */
class OoxLock implements ILock
{
  /**
   * Default arbitrary lifetime for lock. 
   * 
   * @var int
   */
  const LIFETIME_DEFAULT = 30;

  /**
   * Singleton implementation.
   * 
   * @var ILock
   */
  private static $__instance;

  /**
   * Singleton implementation, get instance.
   * 
   * @return ILock
   */
  public static function getInstance() {
    if (!isset(self::$__instance)) {
      // Allow runtime implementation change. Some users might want to use
      // a cache based implementation (using memcache for example could be
      // a lot faster than use database queries).
      if ($class = variable_get(OOX_VAR_LOCK_BACKEND, 0)) {
        if (!oox_class_is($class, 'ILock')) {
          $class = 'OoxLock';
          watchdog('oox', "Defined lock implementation " . $class . " does not implements the ILock interface", NULL, WATCHDOG_ERROR);
        }
        else {
          self::$__instance = new $class();
          if (!self::$__instance->checkEnvironment()) {
            unset(self::$__instance);
            watchdog('oox', "Your environement does not matches the " . $class . " implementation requirements", NULL, WATCHDOG_ERROR);
          }
        }
      }
      // Failsafe, if not instanciated upper, use the default implementation.
      if (!isset(self::$__instance)) {
        self::$__instance = new OoxLock();
      }
    }
    return self::$__instance;
  }

  /**
   * (non-PHPdoc)
   * @see ILock::acquire()
   */
  public function acquire($name, $lifetime = LIFETIME_DEFAULT, $group = ILock::DEFAULT_GROUP, $forUser = FALSE, $token = NULL) {
    $time = time();

    // Aptempt to find if this semaphore already been locked.
    $lock = db_select('semaphore_oox', 's')
      ->fields('s')
      ->condition('expire', $time, '>')
      ->condition('name', $name)
      ->execute()
      ->fetchObject();

    if ($lock) {
      // Success in case of identical tokens.
      if ($token && $lock->token == $token) {
        return TRUE;
      }
      else {
        throw new LockAlreadyAcquiredException($lock);
      }
    }
    else {
      // Avoid collisions when an expired semaphore still exists with the
      // same name.
      db_delete('semaphore_oox')
        ->condition('name', $name)
        ->execute();
    }

    // Get user information.
    if ($forUser) {
      global $user;
      $uid = $user->uid;
    }
    else {
      $uid = 0;
    }

    // If none found, then aptempt the acquisition transaction.
    $transaction = db_transaction();
    try {
      db_insert('semaphore_oox')->fields(array(
        'name' => $name,
        'value' => $group,
        'expire' => $time + $lifetime,
        'uid' => $uid,
        'token' => $token,
      ))->execute();
      // Force transaction to commit.
      unset($transaction);
    }
    // Check for transaction error.
    catch (PDOException $e) {
      $transaction->rollback();
      // Specific case, that should not happens that often, two transactions
      // have been overlapping each other. We should fetch back the current
      // lock and throw the exception we the right user.
      $lock = db_select('semaphore_oox', 's')
        ->fields('s')
        ->condition('expire', $time, '>')
        ->condition('name', $name)
        ->execute()
        ->fetchObject();
      throw new LockAlreadyAcquiredException($lock);
    }
  }

  /**
   * (non-PHPdoc)
   * @see ILock::release()
   */
  public function release($name) {
    $transaction = db_transaction();
    try {
      db_delete('semaphore_oox')
        ->condition('name', $name)
        ->execute();
      // Force transaction to commit.
      unset($transaction);
    }
    // Check for transaction error.
    catch (PDOException $e) {
      $transaction->rollback();
      // This is a real error that should not happen, log why.
      watchdog_exception('oox_lock', $e);
    }
  }

  /**
   * (non-PHPdoc)
   * @see ILock::getLock()
   */
  public function getLock($name, $group = NULL, $uid = NULL, $token = NULL) {
    // Prepare query.
    $select = db_select('semaphore_oox')
      ->fields('s')
      ->condition('expire', time(), '>')
      ->condition('name', $name);
    // Check for given group
    if ($group) {
      $select->condition('value', $group);
    }
    if ($uid) {
      $select->condition('uid', $uid);
    }
    if ($token) {
      $select->condition('token', $token);
    }
    return $select->execute()->fetchObject();
  }

  /**
   * (non-PHPdoc)
   * @see ILock::wipeOutLocks()
   */
  public function wipeOutLocks($uid = 0, $group = NULL, $token = NULL) {
    // Prepare query.
    $delete = db_delete('semaphore_oox');
    $all = TRUE;
    // Check for given group
    if ($group) {
      $delete->condition('value', $group);
      $all = FALSE;
    }
    if ($uid) {
      $delete->condition('uid', $uid);
      $all = FALSE;
    }
    if ($token) {
      $delete->condition('token', $token);
      $all = FALSE;
    }
    // May be this is a maintenance call.
    if ($all) {
      $delete->condition('expire', time(), '<');
    }
    // Do the delete inside a transaction, it could overlap with a lock
    // acquire transaction.
    $transaction = db_transaction();
    try {
      $delete->execute();
      // Force transaction to commit.
      unset($transaction);
    }
    // Check for transaction error.
    catch (PDOException $e) {
      $transaction->rollback();
      // This is a real error that should not happen, log why.
      watchdog_exception('oox_lock', $e);
    }
  }

  /**
   * (non-PHPdoc)
   * @see ILock::checkEnvironment()
   */
  public function checkEnvironment() {
    return TRUE;
  }
}
